# Copyright (c) 2018-2019, NVIDIA CORPORATION. All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os
import sys

JETSON_XAVIER = 'JETSON_XAVIER'
JETSON_TX2 = 'JETSON_TX2'
JETSON_TX1 = 'JETSON_TX1'
JETSON_NANO = 'JETSON_NANO'

# These arrays contain tuples of all the relevant GPIO data for each Jetson
# Platform. The format is represented as so - (Linux GPIO pin number,
# GPIO chip directory, board pin number, BCM pin number, CVM pin number,
# TEGRA_SOC pin number)
# The values are use to generate dictionaries that map the corresponding pin
# mode numbers to the Linux GPIO pin number and GPIO chip directory
JETSON_XAVIER_PIN_DEFS = [
    (134, "/sys/devices/2200000.gpio", 7, 4, 'MCLK05', 'SOC_GPIO42'),
    (140, "/sys/devices/2200000.gpio", 11, 17, 'UART1_RTS', 'UART1_RTS'),
    (63, "/sys/devices/2200000.gpio", 12, 18, 'I2S2_CLK', 'DAP2_SCLK'),
    (136, "/sys/devices/2200000.gpio", 13, 27, 'PWM01', 'SOC_GPIO44'),
    (105, "/sys/devices/2200000.gpio", 15, 22, 'GPIO27', 'SOC_GPIO54'),
    (8, "/sys/devices/c2f0000.gpio", 16, 23, 'GPIO8', 'CAN1_STB'),
    (56, "/sys/devices/2200000.gpio", 18, 24, 'GPIO35', 'SOC_GPIO12'),
    (205, "/sys/devices/2200000.gpio", 19, 10, 'SPI1_MOSI', 'SPI1_MOSI'),
    (204, "/sys/devices/2200000.gpio", 21, 9, 'SPI1_MISO', 'SPI1_MISO'),
    (129, "/sys/devices/2200000.gpio", 22, 25, 'GPIO17', 'SOC_GPIO21'),
    (22, "/sys/devices/2200000.gpio", 23, 11, 'SPI1_CLK', 'SPI1_SCK'),
    (206, "/sys/devices/2200000.gpio", 24, 8, 'SPI1_CS0_N', 'SPI1_CS0_N'),
    (207, "/sys/devices/2200000.gpio", 26, 7, 'SPI1_CS1_N', 'SPI1_CS1_N'),
    (3, "/sys/devices/c2f0000.gpio", 29, 5, 'CAN0_DIN', 'CAN0_DIN'),
    (2, "/sys/devices/c2f0000.gpio", 31, 6, 'CAN0_DOUT', 'CAN0_DOUT'),
    (9, "/sys/devices/c2f0000.gpio", 32, 12, 'GPIO9', 'CAN1_EN'),
    (0, "/sys/devices/c2f0000.gpio", 33, 13, 'CAN1_DOUT', 'CAN1_DOUT'),
    (66, "/sys/devices/2200000.gpio", 35, 19, 'I2S2_FS', 'DAP2_FS'),
    (141, "/sys/devices/2200000.gpio", 36, 16, 'UART1_CTS', 'UART1_CTS'),
    (1, "/sys/devices/c2f0000.gpio", 37, 26, 'CAN1_DIN', 'CAN1_DIN'),
    (65, "/sys/devices/2200000.gpio", 38, 20, 'I2S2_DIN', 'DAP2_DIN'),
    (64, "/sys/devices/2200000.gpio", 40, 21, 'I2S2_DOUT', 'DAP2_DOUT')
]

JETSON_TX2_PIN_DEFS = [
    (76, "/sys/devices/2200000.gpio", 7, 4, 'AUDIO_MCLK', 'AUD_MCLK'),
    (146, "/sys/devices/2200000.gpio", 11, 17, 'UART0_RTS', 'UART1_RTS'),
    (72, "/sys/devices/2200000.gpio", 12, 18, 'I2S0_CLK', 'DAP1_SCLK'),
    (77, "/sys/devices/2200000.gpio", 13, 27, 'GPIO20_AUD_INT', 'GPIO_AUD0'),
    (15, "/sys/devices/3160000.i2c/i2c-0/0-0074", 15, 22, 'GPIO_EXP_P17', 'GPIO_EXP_P17'),
    (40, "/sys/devices/c2f0000.gpio", 16, 23, 'AO_DMIC_IN_DAT', 'CAN0_GPIO0'),
    (161, "/sys/devices/2200000.gpio", 18, 24, 'GPIO16_MDM_WAKE_AP', 'GPIO_MDM2'),
    (109, "/sys/devices/2200000.gpio", 19, 10, 'SPI1_MOSI', 'GPIO_CAM6'),
    (108, "/sys/devices/2200000.gpio", 21, 9, 'SPI1_MISO', 'GPIO_CAM5'),
    (14, "/sys/devices/3160000.i2c/i2c-0/0-0074", 22, 25, 'GPIO_EXP_P16', 'GPIO_EXP_P16'),
    (107, "/sys/devices/2200000.gpio", 23, 11, 'SPI1_CLK', 'GPIO_CAM4'),
    (110, "/sys/devices/2200000.gpio", 24, 8, 'SPI1_CS0', 'GPIO_CAM7'),
    (None, None, 26, 7, 'SPI1_CS1', None),
    (78, "/sys/devices/2200000.gpio", 29, 5, 'GPIO19_AUD_RST', 'GPIO_AUD1'),
    (42, "/sys/devices/c2f0000.gpio", 31, 6, 'GPIO9_MOTION_INT', 'CAN_GPIO2'),
    (41, "/sys/devices/2200000.gpio", 32, 12, 'AO_DMIC_IN_CLK', 'CAN_GPIO1'),
    (69, "/sys/devices/2200000.gpio", 33, 13, 'GPIO11_AP_WAKE_BT', 'GPIO_PQ5'),
    (75, "/sys/devices/2200000.gpio", 35, 19, 'I2S0_LRCLK', 'DAP1_FS'),
    (147, "/sys/devices/2200000.gpio", 36, 16, 'UART0_CTS', 'UART1_CTS'),
    (68, "/sys/devices/2200000.gpio", 37, 26, 'GPIO8_ALS_PROX_INT', 'GPIO_PQ4'),
    (74, "/sys/devices/2200000.gpio", 38, 20, 'I2S0_SDIN', 'DAP1_DIN'),
    (73, "/sys/devices/2200000.gpio", 40, 21, 'I2S0_SDOUT', 'DAP1_DOUT')
]

JETSON_TX1_PIN_DEFS = [
    (216, "/sys/devices/6000d000.gpio", 7, 4, 'AUDIO_MCLK', 'AUD_MCLK'),
    (162, "/sys/devices/6000d000.gpio", 11, 17, 'UART0_RTS', 'UART1_RTS'),
    (11, "/sys/devices/6000d000.gpio", 12, 18, 'I2S0_CLK', 'DAP1_SCLK'),
    (38, "/sys/devices/6000d000.gpio", 13, 27, 'GPIO20_AUD_INT', 'GPIO_PE6'),
    (15, "/sys/devices/7000c400.i2c/i2c-1/1-0074", 15, 22, 'GPIO_EXP_P17', 'GPIO_EXP_P17'),
    (37, "/sys/devices/6000d000.gpio", 16, 23, 'AO_DMIC_IN_DAT', 'DMIC3_DAT'),
    (184, "/sys/devices/6000d000.gpio", 18, 24, 'GPIO16_MDM_WAKE_AP', 'MODEM_WAKE_AP'),
    (16, "/sys/devices/6000d000.gpio", 19, 10, 'SPI1_MOSI', 'SPI1_MOSI'),
    (17, "/sys/devices/6000d000.gpio", 21, 9, 'SPI1_MISO', 'SPI1_MISO'),
    (14, "/sys/devices/7000c400.i2c/i2c-1/1-0074", 22, 25, 'GPIO_EXP_P16', 'GPIO_EXP_P16'),
    (18, "/sys/devices/6000d000.gpio", 23, 11, 'SPI1_CLK', 'SPI1_SCK'),
    (19, "/sys/devices/6000d000.gpio", 24, 8, 'SPI1_CS0', 'SPI1_CS0'),
    (20, "/sys/devices/6000d000.gpio", 26, 7, 'SPI1_CS1', 'SPI1_CS1'),
    (219, "/sys/devices/6000d000.gpio", 29, 5, 'GPIO19_AUD_RST', 'GPIO_X1_AUD'),
    (186, "/sys/devices/6000d000.gpio", 31, 6, 'GPIO9_MOTION_INT', 'MOTION_INT'),
    (36, "/sys/devices/6000d000.gpio", 32, 12, 'AO_DMIC_IN_CLK', 'DMIC3_CLK'),
    (63, "/sys/devices/6000d000.gpio", 33, 13, 'GPIO11_AP_WAKE_BT', 'AP_WAKE_NFC'),
    (8, "/sys/devices/6000d000.gpio", 35, 19, 'I2S0_LRCLK', 'DAP1_FS'),
    (163, "/sys/devices/6000d000.gpio", 36, 16, 'UART0_CTS', 'UART1_CTS'),
    (187, "/sys/devices/6000d000.gpio", 37, 26, 'GPIO8_ALS_PROX_INT', 'ALS_PROX_INT'),
    (9, "/sys/devices/6000d000.gpio", 38, 20, 'I2S0_SDIN', 'DAP1_DIN'),
    (10, "/sys/devices/6000d000.gpio", 40, 21, 'I2S0_SDOUT', 'DAP1_DOUT')
]

JETSON_NANO_PIN_DEFS = [
    (216, "/sys/devices/6000d000.gpio", 7, 4, 'GPIO9', 'GPIO9'),
    (50, "/sys/devices/6000d000.gpio", 11, 17, 'UART1_RTS', 'UART1_RTS'),
    (79, "/sys/devices/6000d000.gpio", 12, 18, 'I2S0_SCLK', 'I2S0_SCLK'),
    (14, "/sys/devices/6000d000.gpio", 13, 27, 'SPI1_SCK', 'SPI1_SCK'),
    (194, "/sys/devices/6000d000.gpio", 15, 22, 'GPIO12', 'GPIO12'),
    (232, "/sys/devices/6000d000.gpio", 16, 23, 'SPI1_CS1', 'SPI1_CS1'),
    (15, "/sys/devices/6000d000.gpio", 18, 24, 'SPI1_CS0', 'SPI1_CS0'),
    (16, "/sys/devices/6000d000.gpio", 19, 10, 'SPI0_MOSI', 'SPI0_MOSI'),
    (17, "/sys/devices/6000d000.gpio", 21, 9, 'SPI0_MISO', 'SPI0_MISO'),
    (13, "/sys/devices/6000d000.gpio", 22, 25, 'SPI1_MISO', 'SPI1_MISO'),
    (18, "/sys/devices/6000d000.gpio", 23, 11, 'SPI0_SCK', 'SPI0_SCK'),
    (19, "/sys/devices/6000d000.gpio", 24, 8, 'SPI0_CS0', 'SPI0_CS0'),
    (20, "/sys/devices/6000d000.gpio", 26, 7, 'SPI0_CS1', 'SPI0_CS1'),
    (149, "/sys/devices/6000d000.gpio", 29, 5, 'GPIO01', 'GPIO01'),
    (200, "/sys/devices/6000d000.gpio", 31, 6, 'GPIO11', 'GPIO11'),
    (168, "/sys/devices/6000d000.gpio", 32, 12, 'GPIO07', 'GPIO07'),
    (38, "/sys/devices/6000d000.gpio", 33, 13, 'GPIO13', 'GPIO13'),
    (76, "/sys/devices/6000d000.gpio", 35, 19, 'I2S0_FS', 'I2S0_FS'),
    (51, "/sys/devices/6000d000.gpio", 36, 16, 'UART1_CTS', 'UART1_CTS'),
    (12, "/sys/devices/6000d000.gpio", 37, 26, 'SPI1_MOSI', 'SPI1_MOSI'),
    (77, "/sys/devices/6000d000.gpio", 38, 20, 'I2S0_DIN', 'I2S0_DIN'),
    (78, "/sys/devices/6000d000.gpio", 40, 21, 'I2S0_DOUT', 'I2S0_DOUT')
]

jetson_gpio_data = {
    'JETSON_XAVIER': {
        'gpio_numbers': {
            'BOARD': {x[2]: (x[0], x[1]) for x in JETSON_XAVIER_PIN_DEFS},
            'BCM': {x[3]: (x[0], x[1]) for x in JETSON_XAVIER_PIN_DEFS},
            'CVM': {x[4]: (x[0], x[1]) for x in JETSON_XAVIER_PIN_DEFS},
            'TEGRA_SOC': {x[5]: (x[0], x[1]) for x in JETSON_XAVIER_PIN_DEFS},
        },
        'JETSON_INFO': {
            'P1_REVISION': 1,
            'RAM': '16384M',
            'REVISION': 'Unknown',
            'TYPE': 'Jetson Xavier',
            'MANUFACTURER': 'NVIDIA',
            'PROCESSOR': 'ARM Carmel'
        }
    },
    'JETSON_TX2': {
        'gpio_numbers': {
            'BOARD': {x[2]: (x[0], x[1]) for x in JETSON_TX2_PIN_DEFS},
            'BCM': {x[3]: (x[0], x[1]) for x in JETSON_TX2_PIN_DEFS},
            'CVM': {x[4]: (x[0], x[1]) for x in JETSON_TX2_PIN_DEFS},
            'TEGRA_SOC': {x[5]: (x[0], x[1]) for x in JETSON_TX2_PIN_DEFS},
        },
        'JETSON_INFO': {
            'P1_REVISION': 1,
            'RAM': '8192M',
            'REVISION': 'Unknown',
            'TYPE': 'Jetson TX2',
            'MANUFACTURER': 'NVIDIA',
            'PROCESSOR': 'ARM A57 + Denver'
        }
    },
    'JETSON_TX1': {
        'gpio_numbers': {
            'BOARD': {x[2]: (x[0], x[1]) for x in JETSON_TX1_PIN_DEFS},
            'BCM': {x[3]: (x[0], x[1]) for x in JETSON_TX1_PIN_DEFS},
            'CVM': {x[4]: (x[0], x[1]) for x in JETSON_TX1_PIN_DEFS},
            'TEGRA_SOC': {x[5]: (x[0], x[1]) for x in JETSON_TX1_PIN_DEFS},
        },
        'JETSON_INFO': {
            'P1_REVISION': 1,
            'RAM': '4096M',
            'REVISION': 'Unknown',
            'TYPE': 'Jetson TX1',
            'MANUFACTURER': 'NVIDIA',
            'PROCESSOR': 'ARM A57'
        }
    },
    'JETSON_NANO': {
        'gpio_numbers': {
            'BOARD': {x[2]: (x[0], x[1]) for x in JETSON_NANO_PIN_DEFS},
            'BCM': {x[3]: (x[0], x[1]) for x in JETSON_NANO_PIN_DEFS},
            'CVM': {x[4]: (x[0], x[1]) for x in JETSON_NANO_PIN_DEFS},
            'TEGRA_SOC': {x[5]: (x[0], x[1]) for x in JETSON_NANO_PIN_DEFS},
        },
        'JETSON_INFO': {
            'P1_REVISION': 1,
            'RAM': '4096M',
            'REVISION': 'Unknown',
            'TYPE': 'Jetson Nano',
            'MANUFACTURER': 'NVIDIA',
            'PROCESSOR': 'ARM A57'
        }
    }
}

def get_gpio_data(model):
    gpio_chip_base = {}
    jetson_model_pin_def = model+"_PIN_DEFS"
    if model not in jetson_gpio_data or model not in jetson_model_pin_def:
        raise Exception("Unknown hardware model " + model)
    gpio_pin_data = jetson_gpio_data[model]
    gpio_chip_dirs = set(filter(None, [x[1] for x in eval(jetson_model_pin_def)]))
    # Get the gpiochip offsets
    for chip in gpio_chip_dirs:
        for filename in os.listdir(chip + '/gpio'):
            if 'gpiochip' in filename:
                with open(chip + '/gpio/' + filename
                          + '/base', 'r') as f:
                    gpio_chip_base[chip] = int(f.read().strip())
                    break
    return gpio_pin_data, gpio_chip_base


# Copyright (c) 2012-2017 Ben Croston <ben@croston.org>.
# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from Jetson.GPIO import gpio_event as event
from Jetson.GPIO import gpio_pin_data
import os
import time
import warnings

# sysfs root
_SYSFS_ROOT = "/sys/class/gpio"

if (not os.access(_SYSFS_ROOT + '/export', os.W_OK) or
        not os.access(_SYSFS_ROOT + '/unexport', os.W_OK)):
    raise RuntimeError("The current user does not have permissions set to "
                       "access the library functionalites. Please configure "
                       "permissions or use the root user to run this")

# Pin Numbering Modes
BOARD = 'BOARD'
BCM = 'BCM'
TEGRA_SOC = 'TEGRA_SOC'
CVM = 'CVM'
_MODE_UNKNOWN = None

# The constants and their offsets are implemented to prevent HIGH from being used
# in place of other variables (ie. HIGH and RISING should not be interchangeable)

# Pull up/down options
PUD_OFF = 20
PUD_DOWN = 21
PUD_UP = 22

HIGH = 1
LOW = 0

# Edge possibilities
RISING = 11
FALLING = 12
BOTH = 13

# Pullup/down and Edge Offset
_PUD_OFFSET = -20
_EDGE_OFFSET = -10

# GPIO directions. UNEXPORTED constant is for gpios that are not yet setup
_UNEXPORTED = ''
OUT = 'out'
IN = 'in'

# check jetson model and prepare lookup table accordingly
def get_model(model_path='/proc/device-tree/model'):
    version_path = '/proc/device-tree/chosen/plugin-manager/ids'
    model_str = None

    try:
        with open(model_path, 'r') as f:
            model_str = f.read().rstrip('\x00')
    except:
        raise Exception('Could not determine Jetson model because model file'
                        '(%s) was not found.' % model_path)

    if 'tx1' in model_str:
        return gpio_pin_data.JETSON_TX1
    elif 'tx2' in model_str or 'quill' in model_str:
        return gpio_pin_data.JETSON_TX2
    elif 'xavier' in model_str:
        return gpio_pin_data.JETSON_XAVIER
    elif 'nano' in model_str:
        if int(os.listdir(version_path)[0][-3:]) >= 200:
            return gpio_pin_data.JETSON_NANO
        else:
            raise Exception('Jetson Nano revision must be newer than A02')

    raise Exception('Could not guess Jetson model from the model string (%s).'
                    % model_str)

_board_info, _gpio_chip_base = gpio_pin_data.get_gpio_data(get_model())
_pin_mapping = _board_info['gpio_numbers']
JETSON_INFO = _board_info['JETSON_INFO']
RPI_INFO = JETSON_INFO

# Dicitionary objects used as lookup tables for pin to linux gpio mapping
_pin_to_gpio = {}

_gpio_warnings = True
_gpio_mode = _MODE_UNKNOWN
_gpio_direction = {}

# Function used to enable/disable warnings during setup and cleanup.
# Param -> state is a bool
def setwarnings(state):
    global _gpio_warnings
    _gpio_warnings = bool(state)

# Function used to set the pin mumbering mode. Possible mode values are BOARD,
# BCM, TEGRA_SOC and CVM
def setmode(mode):
    global _gpio_mode, _pin_to_gpio

    # check if a different mode has been set
    if _gpio_mode != _MODE_UNKNOWN and mode != _gpio_mode:
        raise ValueError("A different mode has already been set!")

    # check if mode parameter is valid
    if mode != BOARD and mode != BCM and mode != TEGRA_SOC and mode != CVM:
        raise ValueError("An invalid mode was passed to setmode()!")

    _pin_to_gpio = _pin_mapping[mode]
    _gpio_mode = mode

# Function used to get the currently set pin numbering mode
def getmode():
    return _gpio_mode

# Function used to setup individual pins or lists/tuples of pins as
# Input or Output. Param channels must an integer or list/tuple of integers,
# direction must be IN or OUT, pull_up_down must be PUD_OFF, PUD_UP or
# PUD_DOWN and is only valid when direction in IN, initial must be HIGH or LOW
# and is only valid when direction is OUT
def setup(channels, direction, pull_up_down=PUD_OFF, initial=None):
    # check if channels are iterable
    channels = _make_iterable(channels)

    if _gpio_mode == BCM or _gpio_mode == BOARD:
        # check if all values in the iterable is an int
        if not all(isinstance(x, int) for x in channels):
            raise ValueError("Channel must be an integer or list/tuple of"
                              "integers")
    elif _gpio_mode == TEGRA_SOC or _gpio_mode == CVM:
        # check if all values in the iterable is a string
        if not all(isinstance(x, str) for x in channels):
            raise ValueError("Channel must be an string or list/tuple of"
                             "strings")

    # check direction is valid
    if direction != OUT and direction != IN:
        raise ValueError("An invalid direction was passed to setup()")

    # check if pullup/down is used with output
    if direction == OUT and pull_up_down != PUD_OFF:
        raise ValueError("pull_up_down parameter is not valid for outputs")

    # check if intial value is used with input
    if direction == IN and initial is not None:
        raise ValueError("initial parameter is not valid for inputs")

    # check if pullup/down value is valid
    if (pull_up_down != PUD_OFF and pull_up_down != PUD_UP and
            pull_up_down != PUD_DOWN):
        raise ValueError("Invalid value for pull_up_down - should be either"
                         "PUD_OFF, PUD_UP or PUD_DOWN")

    if direction == OUT:
        for idx in range(len(channels)):
            if isinstance(initial, (list, tuple)):
                if len(channels) != len(initial):
                    raise RuntimeError("Number of channels != number of "
                                       "initial values")
                _setup_single_out(channels[idx], initial[idx])
            else:
                _setup_single_out(channels[idx], initial)
    else:
        for idx in range(len(channels)):
            if isinstance(initial, (list, tuple)):
                if len(channels) != len(initial):
                    raise RuntimeError("Number of channels != number of "
                                       "initial values")
                _setup_single_in(channels[idx], pull_up_down + _PUD_OFFSET)
            else:
                _setup_single_in(channels[idx], pull_up_down + _PUD_OFFSET)


def _setup_single_out(channel, initial=None):
    gpio = _get_gpio_number(channel)
    func = gpio_function(channel)
    direction = _check_pin_setup(gpio)

    # warn if channel has been setup external to currently running program
    if _gpio_warnings and direction is None:
        if func == OUT or func == IN:
            warnings.warn("This channel is already in use, continuing anyway. "
                          "Use GPIO.setwarnings(False) to disable warnings",
                          RuntimeWarning)

    _export_gpio(gpio)
    gpio_dir_path = _SYSFS_ROOT + "/gpio%i" % gpio + "/direction"
    gpio_value_path = _SYSFS_ROOT + "/gpio%i" % gpio + "/value"

    with open(gpio_dir_path, 'w') as direction_file:
        direction_file.write(OUT)

    if initial == HIGH or initial == LOW:
        with open(gpio_value_path, 'w') as value:
            initial = int(initial)
            value.write(str(initial))

    _gpio_direction[gpio] = OUT

def _setup_single_in(channel, pull_up_down=PUD_OFF):
    gpio = _get_gpio_number(channel)
    func = gpio_function(channel)
    direction = _check_pin_setup(gpio)

    # warn if channel has been setup external to currently running program
    if _gpio_warnings and direction is None:
        if func == OUT or func == IN:
            warnings.warn("This channel is already in use, continuing anyway. "
                          "Use GPIO.setwarnings(False) to disable warnings",
                          RuntimeWarning)

    _export_gpio(gpio)
    gpio_dir_path = _SYSFS_ROOT + "/gpio%i" % gpio + "/direction"

    with open(gpio_dir_path, 'w') as direction:
        direction.write(IN)

    _gpio_direction[gpio] = IN

# Function used to cleanup channels at the end of the program.
# The param channel can be an integer or list/tuple of integers specifying the
# channels to be cleaned up. If no channel is provided, all channels are cleaned
def cleanup(channel=None):
    # warn if no channel is setup
    if _gpio_mode == _MODE_UNKNOWN:
        if _gpio_warnings:
            warnings.warn("No channels have been set up yet - nothing to "
                          "clean up! Try cleaning up at the end of your "
                          "program instead!", RuntimeWarning)
        return

    # clean all channels if no channel param provided
    if channel is None:
        _cleanup_all()
        return

    # check if channels are iterable
    channel = _make_iterable(channel)

    if all(isinstance(x, (str, int)) for x in channel):
        for idx in channel:
            _cleanup_one(idx)

    else:
        raise ValueError("Channel must be an integer/string or list/tuple of "
                         "integers/strings")

def _cleanup_one(channel):
    gpio = _get_gpio_number(channel)
    if gpio is not None or _check_pin_setup(gpio) is not None:
        _setup_single_in(channel)
        del _gpio_direction[gpio]
        event.event_cleanup(gpio)
        _unexport_gpio(gpio)

def _cleanup_all():
    global _gpio_mode

    for channel in _pin_to_gpio:
        if _pin_to_gpio[channel][0] is not None:
            _cleanup_one(channel)

    _gpio_mode = _MODE_UNKNOWN

# Function used to return the current value of the specified channel.
# Function returns either HIGH or LOW
def input(channel):
    gpio = _get_gpio_number(channel)
    direction = _check_pin_setup(gpio)
    if direction != IN and direction != OUT:
        raise RuntimeError("You must setup() the GPIO channel first")

    with open(_SYSFS_ROOT + "/gpio%i" % gpio + "/value") as value:
        value_read = int(value.read())
        return value_read

# Function used to set a value to a channel or list/tuple of channels.
# Parameter channels must be an integer or list/tuple of integers.
# Values must be either HIGH or LOW or list/tuple
# of HIGH and LOW with the same length as the channels list/tuple
def output(channels, values):
    # check if channels and values are iterable
    channels = _make_iterable(channels)
    values = _make_iterable(values)

    if _gpio_mode == BOARD or _gpio_mode == BCM:
        # check if all elements in the channels iterable are integers
        if not all(isinstance(x, int) for x in channels):
            raise ValueError("Channel must be an integer or list/tuple of "
                             "integers")
    elif _gpio_mode == TEGRA_SOC or _gpio_mode == CVM:
        # check if all elements in the channels iterable are strings
        if not all(isinstance(x, str) for x in channels):
            raise ValueError("Channel must be a string or "
                             "list/tuple of string")

    # check if all elements in the values iterable are integers/booleans
    if not all(isinstance(x, (int, bool)) for x in values):
        raise RuntimeError("Value list must consist of integers/booleans")

    # check that channels have been set as output
    if any((_check_pin_setup(_get_gpio_number(x)) != OUT)
           for x in channels):
        raise RuntimeError("The GPIO channel has not been set up as an "
                           "OUTPUT")

    # check if values contains more than 1 element and is not empty
    if values:
        try:
            values[1]
            if len(channels) != len(values):
                raise RuntimeError("Number of channels != number of values")
            else:
                for idx in range(len(channels)):
                    _output_one(channels[idx], values[idx])
        except:
            for idx in range(len(channels)):
                _output_one(channels[idx], values[0])
    else:
        raise ValueError("Value must not be empty")

def _check_pin_setup(gpio):
    return _gpio_direction.get(gpio, None)

def _make_iterable(iterable):
    try:
        for x in iterable:
            break
    except:
        iterable = [iterable]

    if isinstance(iterable, str):
        iterable = [iterable]

    return iterable

def _output_one(channel, value):
    gpio = _get_gpio_number(channel)
    value = int(value)
    if value != HIGH and value != LOW:
        raise ValueError("Invalid output value")

    with open(_SYSFS_ROOT + "/gpio%s" % gpio + "/value", 'w') as value_file:
        value = str(value)
        value_file.write(value)

# Function used to check if an event occurred on the specified channel.
# Param channel must be an integer.
# This function return True or False
def event_detected(channel):
    gpio = _get_gpio_number(channel)
    return event.edge_event_detected(gpio)

# Function used to add a callback function to channel, after it has been
# registered for events using add_event_detect()
def add_event_callback(channel, callback):
    if not callable(callback):
        raise TypeError("Parameter must be callable")

    gpio = _get_gpio_number(channel)
    if _check_pin_setup(gpio) != IN:
        raise RuntimeError("You must setup() the GPIO channel as an "
                           "input first")

    if not event.gpio_event_added(gpio):
        raise RuntimeError("Add event detection using add_event_detect first "
                           "before adding a callback")

    event.add_edge_callback(gpio, lambda: callback(channel))

# Function used to add threaded event detection for a specified gpio channel.
# Param gpio must be an integer specifying the channel, edge must be RISING,
# FALLING or BOTH. A callback function to be called when the event is detected
# and an integer bounctime in milliseconds can be optionally provided
def add_event_detect(channel, edge, callback=None, bouncetime=None):
    result = None

    if (not callable(callback)) and callback is not None:
        raise TypeError("Callback Parameter must be callable")

    gpio = _get_gpio_number(channel)

    # channel must be setup as input
    if _check_pin_setup(gpio) != IN:
        raise RuntimeError("You must setup() the GPIO channel as an input "
                           "first")

    # edge must be rising, falling or both
    if edge != RISING and edge != FALLING and edge != BOTH:
        raise ValueError("The edge must be set to RISING, FALLING, or BOTH")

    # if bouncetime is provided, it must be int and greater than 0
    if bouncetime is not None:
        if type(bouncetime) != int:
            raise TypeError("bouncetime must be an integer")

        elif bouncetime < 0:
            raise ValueError("bouncetime must be an integer greater than 0")

    result = event.add_edge_detect(gpio, edge + _EDGE_OFFSET, bouncetime)

    # result == 1 means a different edge was already added for the channel.
    # result == 2 means error occurred while adding edge (thread or event poll)
    if result:
        error_str = None
        if result == 1:
            error_str = "Conflicting edge already enabled for this GPIO channel"
        else:
            error_str = "Failed to add edge detection"

        raise RuntimeError(error_str)

    if callback is not None:
        event.add_edge_callback(gpio, lambda: callback(channel))

# Function used to remove event detection for channel
def remove_event_detect(channel):
    gpio = _get_gpio_number(channel)
    event.remove_edge_detect(gpio)

# Function used to perform a blocking wait until the specified edge
# is detected for the param channel. Channel must be an integer and edge must
# be either RISING, FALLING or BOTH.
# bouncetime in milliseconds and timeout in millseconds can optionally be
# provided
def wait_for_edge(channel, edge, bouncetime=None, timeout=None):
    gpio = _get_gpio_number(channel)

    # channel must be setup as input
    if _check_pin_setup(gpio) != IN:
        raise RuntimeError("You must setup() the GPIO channel as an input "
                           "first")

    # edge provided must be rising, falling or both
    if edge != RISING and edge != FALLING and edge != BOTH:
        raise ValueError("The edge must be set to RISING, FALLING_EDGE "
                         "or BOTH")

    # if bouncetime is provided, it must be int and greater than 0
    if bouncetime is not None:
        if type(bouncetime) != int:
            raise TypeError("bouncetime must be an integer")

        elif bouncetime < 0:
            raise ValueError("bouncetime must be an integer greater than 0")

    # if timeout is specified, it must be an int and greater than 0
    if timeout is not None:
        if type(timeout) != int:
            raise TypeError("Timeout must be an integer")

        elif timeout < 0:
            raise ValueError("Timeout must greater than 0")

    result = event.blocking_wait_for_edge(gpio, edge + _EDGE_OFFSET,
                                          bouncetime, timeout)

    # If not error, result == channel. If timeout occurs while waiting,
    # result == None. If error occurs, result == -1 means channel is
    # registered for conflicting edge detection, result == -2 means an error
    # occurred while registering event or polling
    if not result:
        return None
    elif result == -1:
        raise RuntimeError("Conflicting edge detection event already exists "
                           "for this GPIO channel")

    elif result == -2:
        raise RuntimeError("Error waiting for edge")

    else:
        return channel

def _get_gpio_number(channel):
    if (_gpio_mode != BOARD and _gpio_mode != BCM and
            _gpio_mode != TEGRA_SOC and _gpio_mode != CVM):
        raise RuntimeError("Please set pin numbering mode using "
                           "GPIO.setmode(GPIO.BOARD), GPIO.setmode(GPIO.BCM), "
                           "GPIO.setmode(GPIO.TEGRA_SOC) or "
                           "GPIO.setmode(GPIO.CVM)")

    if channel not in _pin_to_gpio or _pin_to_gpio[channel][0] is None:
        raise ValueError("The channel sent is invalid on a {} Board".format(
            get_model()))

    return _pin_to_gpio[channel][0] + _gpio_chip_base[_pin_to_gpio[channel][1]]

# Function used to check the currently set function of the channel specified.
# Param channel must be an integers. The function returns either IN, OUT,
# or UNEXPORTED
def gpio_function(channel):
    gpio = _get_gpio_number(channel)
    gpio_dir = _SYSFS_ROOT + "/gpio%i" % gpio

    if not os.path.exists(gpio_dir):
        return _UNEXPORTED

    with open(gpio_dir + "/direction", 'r') as f_direction:
        function_ = f_direction.read()

    return function_.rstrip()

def _export_gpio(gpio):
    if os.path.exists(_SYSFS_ROOT + "/gpio%i" % gpio):
        return

    with open(_SYSFS_ROOT + "/export", "w") as f_export:
        f_export.write(str(gpio))

    while not os.access(_SYSFS_ROOT + "/gpio%i" %gpio + "/direction",
                        os.R_OK | os.W_OK):
        time.sleep(0.01)

def _unexport_gpio(gpio):
    if not os.path.exists(_SYSFS_ROOT + "/gpio%i" % gpio):
        return

    with open(_SYSFS_ROOT + "/unexport", "w") as f_unexport:
        f_unexport.write(str(gpio))